from typing import Optional, List
from collections import UserList

import pydantic

from .error import LicenseNotFoundError, LicenseOriginURLMissingError, MultipleLicensesWarning

class License(pydantic.BaseModel):
    direct_url: Optional[str]  # radio optional <a> (and url type)
    xml_url: Optional[str]  # radio optional <a> (and url type)
    full_name: str
    spdx_id: str
    approved_osi: bool
    approved_fsf: bool
    license_header_text: Optional[str]  # format region
    license_text: str
    obsolete: bool

    @pydantic.root_validator(pre=True)
    @classmethod
    def check_direct_or_xml_url_available(cls, values):
        if not ("direct_url" in values or "xml_url" in values):
            raise LicenseOriginURLMissingError(values)
        return values

class Licenses(UserList):
    # convert to list and from list init and more
    def __init__(self, licenses:List[License]):
        self.data = licenses
    
    def get(self, name):
        pass
    
    def filter(self, prop):
        pass
    
    def __getitem__(self, name): # only one and exact
        # change filter to get
        # check all name and see you got one or more as test
        """When using getitem to get the license, you have to enter exact
        spdx_id or full_name. #filter

        Args:
            name (_type_): _description_

        Returns:
            _type_: _description_
        """
        result = list(filter(
            lambda license:
                (license.full_name == name) or
                (license.spdx_id == name),
                self.data))
        if len(result) == 0:
            raise LicenseNotFoundError(name)
        elif len(result) == 1:
            return result[0]
        else:
            MultipleLicensesWarning(name)
            return result
        
    # def __getattr__(self, name): # more than one? with search
    #     return get()



from whoosh.index import create_in
from whoosh.fields import *
from whoosh.qparser import QueryParser

li = ["javad", "kaser", "kasem", "raddi", "GPL3", "GPLv3", "GPLv3.0_only", "GPLv3.0_and_next"]

# search in li for a string with whoosh
schema = Schema(title=TEXT(stored=True))
ix = create_in("indexdir", schema)

writer = ix.writer()
for item in li:
    writer.add_document(title=item)    
writer.commit()

with ix.searcher(score = sco/) as searcher:
    query = QueryParser("title", ix.schema).parse("GPLve3")
    results = searcher.search(query)
    print(results.top_n)


# parse "https://spdx.org/licenses/" and get the table with sortable class beautifulsoup4

from pathlib import Path
from enum import Enum
from typing import List, Union, IO
import json
import importlib.resources as pkg_resources

from .type import License, Licenses


def dump_licenses(
    licenses: List[License], filepath: Path = Path("./licenses.json"), join: bool = True
):  # join do nothing
    with open(str(filepath.resolve()), "w", encoding="utf-8") as f:
        licenses = [license.dict() for license in licenses]
        json.dump(licenses, f, ensure_ascii=False, indent=4)

def load_licenses(licenses:Union[str, Path, IO[str], IO[bytes]] = Path("./licenses.json")):
    if (licenses is str):
        pass
    if (licenses is Path):
        licenses = licenses.resolve().read_text()
    if (licenses is IO[str]):
        licenses = licenses.read() # test
    if (licenses is IO[bytes]):
        licenses = licenses.read().decode("utf-8")  # test
    licenses = json.loads(licenses) # convert to licenses data type and each one a license class
    return Licenses([License(**license) for license in licenses])

def load_builtin_licenses():
    from . import data
    licenses = pkg_resources.read_text(data, "licenses_nonobsolete_only.json")
    return load_licenses(licenses)


import warnings

class LicenseOriginURLMissingError(Exception):
    def __init__(self, values) -> None:
        super().__init__(
            f'License "{values.full_name}" missing any URL pointing to its origin'
        )

class LicenseNotFoundError(ValueError):
    def __init__(self, name) -> None:
        super().__init__(
            f'License with spdx_id or full_name "{name}" is not available'
        )
        
def MultipleLicensesWarning(name):
    warnings.warn("Multiple licenses with spdx_id or full_name {} are available, returning a List".format(name), RuntimeWarning)


from typing import List
import requests
import asyncio
import os

import aiohttp
from bs4 import BeautifulSoup

from .type import License, Licenses

if os.name == "nt":
    print("windows")
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

spdx_url_direct = "https://spdx.org/licenses/"
spdx_url_xmls = ""  # not determined yet

# Lines with "#danger" at the end are dangerous and in case the website template
# update, they may break the code, check them in case of malfunction first.


def spdx_scrapper_direct(
    log: bool = False,
) -> Licenses:  # make it with async requests
    response = requests.get(spdx_url_direct)
    response.raise_for_status()

    soup = BeautifulSoup(response.text, "html.parser")
    table = soup.find(
        "table", {"class": "sortable"}
    )  # also unsotable one append to this table
    session = aiohttp.ClientSession()
    licenses = []
    for i, row in enumerate(table.find_all("tr")):
        if row.find("th"):
            continue
        licenses.append(__get_license_header_and_text(session, row, i, log))
    licenses = asyncio.gather(*licenses)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(licenses)
    licenses = licenses.result()
    loop.run_until_complete(session.close())
    loop.close()
    return Licenses(licenses)
    # return asyncio.gather(
    #     # iterate over table rows in order to move those with either OSI or FSF approve to a sep dictinary
    #     *[
    #         __get_license_header_and_text(session, row)
    #         for row in table.find_all("tr")
    #         if not row.find("th")
    #     ]#, return_exceptions = True # wtf is this?
    # )


async def __get_license_header_and_text(session, row, i, log):
    full_name, spdx_id, approved_osi, approved_fsf = row.find_all("td")
    url = (
        spdx_url_direct + full_name.find("a").get("href")[2:]
    )  # danger (also because of find_all order is not promised)
    full_name = full_name.find("a").text.strip()
    spdx_id = spdx_id.find("code").text.strip()
    approved_osi = approved_osi.text.strip() == "Y"  # danger
    approved_fsf = approved_fsf.text.strip() == "Y"  # danger
    async with session.get(url) as response:
        soup = BeautifulSoup(await response.text(), "html.parser")
        license_header_text = "".join(
            soup.find("div", {"property": "spdx:standardLicenseHeader"}).findAll(
                text=True
            )
        ).strip()
        if license_header_text == "There is no standard license header for the license":
            license_header_text = None
        license_text = "".join(
            soup.find("div", {"property": "spdx:licenseText"}).findAll(text=True)
        ).strip()
        if log:
            print(f"{i} done: {spdx_id}")
        return License(
            direct_url=url,
            full_name=full_name,
            spdx_id=spdx_id,
            approved_osi=approved_osi,
            approved_fsf=approved_fsf,
            license_header_text=license_header_text,
            license_text=license_text,
            obsolete=False,
        )


# def spdx_scrapper_xmls(log:bool=False) -> List[License]:
#     pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['src']

package_data = \
{'': ['*']}

modules = \
['licenses_nonobsolete_only']
install_requires = \
['pydantic>=1.9.1,<2.0.0']

setup_kwargs = {
    'name': 'licenses-spdx',
    'version': '0.1.0',
    'description': 'A powerful API to grab spdx list licenses, great for boilerplate gen and way more (not official)',
    'long_description': None,
    'author': 'aarmn',
    'author_email': 'aarmn80@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
